"""Metadata package for libvcs."""

from __future__ import annotations

__title__ = "libvcs"
__package_name__ = "libvcs"
__description__ = "Lite, typed, python utilities for Git, SVN, Mercurial, etc."
__version__ = "0.45.1"
__author__ = "Tony Narlock"
__github__ = "https://github.com/vcs-python/libvcs"
__docs__ = "https://libvcs.git-pull.com"
__tracker__ = "https://github.com/vcs-python/libvcs/issues"
__pypi__ = "https://pypi.org/project/libvcs/"
__email__ = "tony@git-pull.com"
__license__ = "MIT"
__copyright__ = "Copyright 2016- Tony Narlock"



"""Detect, parse, and validate SVN (Subversion) URLs.

- Detect: :meth:`SvnURL.is_valid()`
- Parse: :class:`SvnURL`

  compare to :class:`urllib.parse.ParseResult`

  - Output ``svn(1)`` URL: :meth:`SvnURL.to_url()`
- Extendable via :class:`~libvcs.url.base.RuleMap`,
  :class:`~libvcs.url.base.Rule`

.. Note::

   Subversion URL support is experimental. If you rely on these URL shapes in
   production, please verify them against your repositories and report gaps in
   the `project tracker <https://github.com/vcs-python/libvcs>`_.
"""

from __future__ import annotations

import dataclasses
import re

from libvcs._internal.dataclasses import SkipDefaultFieldsReprMixin

from .base import Rule, RuleMap, URLProtocol
from .constants import RE_PIP_REV, RE_SCP, RE_USER

RE_PATH = r"""
    (?P<hostname>([^/:@]+))
    (:(?P<port>\d{1,5}))?
    (?P<separator>[:,/])?
    (?P<path>
      (\w[^:.@]*)
    )?
"""

# Valid schemes for svn(1).
# See Table 1.1 Repository access URLs in SVN Book
# https://svnbook.red-bean.com/nightly/en/svn.basic.in-action.html#svn.basic.in-action.wc.tbl-1
RE_SCHEME = r"""
    (?P<scheme>
      (
        file|http|https|svn|svn\+ssh
      )
    )
"""

DEFAULT_RULES: list[Rule] = [
    Rule(
        label="core-svn",
        description="Vanilla svn pattern",
        pattern=re.compile(
            rf"""
        ^{RE_SCHEME}
        ://
        {RE_USER}
        {RE_PATH}
        {RE_PIP_REV}?
        """,
            re.VERBOSE,
        ),
    ),
    Rule(
        label="core-svn-scp",
        description="Vanilla scp(1) / ssh(1) type URL",
        pattern=re.compile(
            rf"""
        ^(?P<scheme>ssh)?
        {RE_USER}
        {RE_SCP}
        {RE_PIP_REV}?
        """,
            re.VERBOSE,
        ),
        defaults={"username": "svn"},
    ),
    # SCP-style URLs, e.g. hg@
]
"""Core regular expressions. These are patterns understood by ``svn(1)``"""


#
# Third-party URLs, e.g. npm, pip, etc.
#
RE_PIP_SCHEME = r"""
    (?P<scheme>
      (
        svn\+ssh|
        svn\+https|
        svn\+http
      )
    )
"""

PIP_DEFAULT_RULES: list[Rule] = [
    Rule(
        label="pip-url",
        description="pip-style svn URL",
        pattern=re.compile(
            rf"""
        ^{RE_PIP_SCHEME}
        ://
        {RE_USER}
        {RE_PATH}
        {RE_PIP_REV}?
        """,
            re.VERBOSE,
        ),
        is_explicit=True,
    ),
    # file://, RTC 8089, File:// https://datatracker.ietf.org/doc/html/rfc8089
    Rule(
        label="pip-file-url",
        description="pip-style svn+file:// URL",
        pattern=re.compile(
            r"""
        (?P<scheme>svn\+file)://
        (?P<path>.*)
        """,
            re.VERBOSE,
        ),
        is_explicit=True,
    ),
]
"""pip-style svn URLs.

Examples of PIP-style svn URLs (via pip.pypa.io)::

    MyProject @ svn+https://svn.example.com/MyProject
    MyProject @ svn+ssh://svn.example.com/MyProject
    MyProject @ svn+ssh://user@svn.example.com/MyProject

Refs (via pip.pypa.io)::

    MyProject @ -e svn+http://svn.example.com/svn/MyProject/trunk@2019
    MyProject @ -e svn+http://svn.example.com/svn/MyProject/trunk@{20080101}

Notes
-----

- https://pip.pypa.io/en/stable/topics/vcs-support/
"""


@dataclasses.dataclass(repr=False)
class SvnBaseURL(
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """SVN repository location. Parses URLs on initialization.

    Examples
    --------
    >>> SvnBaseURL(
    ...     url='svn+ssh://svn.debian.org/svn/aliothproj/path/in/project/repository')
    SvnBaseURL(url=svn+ssh://svn.debian.org/svn/aliothproj/path/in/project/repository,
           scheme=svn+ssh,
           hostname=svn.debian.org,
           path=svn/aliothproj/path/in/project/repository,
           rule=core-svn)

    >>> myrepo = SvnBaseURL(
    ...     url='svn+ssh://svn.debian.org/svn/aliothproj/path/in/project/repository'
    ... )

    >>> myrepo.hostname
    'svn.debian.org'

    >>> myrepo.path
    'svn/aliothproj/path/in/project/repository'

    - Compatibility checking: :meth:`SvnBaseURL.is_valid()`
    - URLs compatible with ``svn(1)``: :meth:`SvnBaseURL.to_url()`
    """

    url: str
    scheme: str | None = None
    user: str | None = None
    hostname: str = dataclasses.field(default="")
    port: int | None = None
    separator: str = dataclasses.field(default="/")
    path: str = dataclasses.field(default="")

    ref: str | None = None

    rule: str | None = None
    rule_map = RuleMap(_rule_map={m.label: m for m in DEFAULT_RULES})

    def __post_init__(self) -> None:
        """Initialize SvnURL params into attributes."""
        url = self.url
        for rule in self.rule_map.values():
            match = re.match(rule.pattern, url)
            if match is None:
                continue
            groups = match.groupdict()
            self.rule = rule.label
            for k, v in groups.items():
                setattr(self, k, v)

            for k in rule.defaults:
                if getattr(self, k, None) is None:
                    setattr(self, k, rule.defaults[k])

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        """Whether URL is compatible with VCS or not.

        Examples
        --------
        >>> SvnBaseURL.is_valid(
        ...     url='svn+ssh://svn.debian.org/svn/aliothproj/path/in/project/repository'
        ... )
        True

        >>> SvnBaseURL.is_valid(url='notaurl')
        False
        """
        if is_explicit is not None:
            return any(
                re.search(rule.pattern, url)
                for rule in cls.rule_map.values()
                if rule.is_explicit == is_explicit
            )
        return any(re.search(rule.pattern, url) for rule in cls.rule_map.values())

    def to_url(self) -> str:
        """Return a ``svn(1)``-compatible URL. Can be used with ``svn checkout``.

        Examples
        --------
        >>> svn_url = SvnBaseURL(
        ...     url='svn+ssh://my-username@my-server/vcs-python/libvcs'
        ... )

        >>> svn_url
        SvnBaseURL(url=svn+ssh://my-username@my-server/vcs-python/libvcs,
                scheme=svn+ssh,
                user=my-username,
                hostname=my-server,
                path=vcs-python/libvcs,
                rule=core-svn)

        Switch repo libvcs -> vcspull:

        >>> svn_url.path = 'vcs-python/vcspull'

        >>> svn_url.to_url()
        'svn+ssh://my-username@my-server/vcs-python/vcspull'

        Switch user to "tom":

        >>> svn_url.user = 'tom'

        >>> svn_url.to_url()
        'svn+ssh://tom@my-server/vcs-python/vcspull'
        """
        if self.scheme is not None:
            parts = [self.scheme, "://"]

            if self.user is not None:
                parts.append(f"{self.user}@")
            parts.append(self.hostname)
        else:
            parts = [self.user or "hg", "@", self.hostname]

        if self.port is not None:
            parts.extend([":", f"{self.port}"])

        parts.extend([self.separator, self.path])

        return "".join(part for part in parts if isinstance(part, str))


@dataclasses.dataclass(repr=False)
class SvnPipURL(
    SvnBaseURL,
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """Supports pip svn URLs."""

    # commit-ish (rev): tag, branch, ref
    rev: str | None = None

    rule_map = RuleMap(_rule_map={m.label: m for m in PIP_DEFAULT_RULES})

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        """Whether URL is compatible with VCS or not.

        Examples
        --------
        >>> SvnPipURL.is_valid(
        ...     url='svn+https://svn.project.org/project-central'
        ... )
        True

        >>> SvnPipURL.is_valid(url='svn+ssh://svn@svn.python.org:cpython')
        True

        >>> SvnPipURL.is_valid(url='notaurl')
        False
        """
        return super().is_valid(url=url, is_explicit=is_explicit)

    def to_url(self) -> str:
        """Return a ``svn(1)``-compatible URL. Can be used with ``svn checkout``.

        Examples
        --------
        >>> svn_url = SvnPipURL(url='svn+https://svn.project.org/project-central')

        >>> svn_url
        SvnPipURL(url=svn+https://svn.project.org/project-central,
                scheme=svn+https,
                hostname=svn.project.org,
                path=project-central,
                rule=pip-url)

        Switch repo project-central -> mobile-browser:

        >>> svn_url.path = 'mobile-browser'

        >>> svn_url.to_url()
        'svn+https://svn.project.org/mobile-browser'

        Switch them to localhost:

        >>> svn_url.hostname = 'localhost'
        >>> svn_url.scheme = 'http'

        >>> svn_url.to_url()
        'http://localhost/mobile-browser'

        """
        return super().to_url()


@dataclasses.dataclass(repr=False)
class SvnURL(
    SvnPipURL,
    SvnBaseURL,
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """Batteries included URL Parser. Supports svn(1) and pip URLs.

    **Ancestors (MRO)**
    This URL parser inherits methods and attributes from the following parsers:

    - :class:`SvnPipURL`

      - :meth:`SvnPipURL.to_url`
    - :class:`SvnBaseURL`

      - :meth:`SvnBaseURL.to_url`
    """

    rule_map = RuleMap(
        _rule_map={m.label: m for m in [*DEFAULT_RULES, *PIP_DEFAULT_RULES]},
    )

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        r"""Whether URL is compatible included Svn URL rule_map or not.

        Examples
        --------
        **Will** match normal ``svn(1)`` URLs, use :meth:`SvnURL.is_valid` for that.

        >>> SvnURL.is_valid(
        ... url='https://svn.project.org/project-central/project-central')
        True

        >>> SvnURL.is_valid(url='svn@svn.project.org:MyProject/project')
        True

        Pip-style URLs:

        >>> SvnURL.is_valid(url='svn+https://svn.project.org/project-central/project')
        True

        >>> SvnURL.is_valid(url='svn+ssh://svn@svn.project.org:MyProject/project')
        True

        >>> SvnURL.is_valid(url='notaurl')
        False

        **Explicit VCS detection**

        Pip-style URLs are prefixed with the VCS name in front, so its rule_map can
        unambiguously narrow the type of VCS:

        >>> SvnURL.is_valid(
        ...     url='svn+ssh://svn@svn.project.org:project-central/image',
        ...     is_explicit=True
        ... )
        True

        Below, while it's svn.project.org, that doesn't necessarily mean that the URL
        itself is conclusively a `svn` URL (e.g. the pattern is too broad):

        >>> SvnURL.is_valid(
        ...     url='svn@svn.project.org:project-central/image', is_explicit=True
        ... )
        False

        You could create a project rule that consider svn.project.org hostnames to be
        exclusively svn:

        >>> projectRule = Rule(
        ...     # Since svn.project.org exclusively serves svn repos, make explicit
        ...     label='project-rule',
        ...     description='Matches svn.project.org https URLs, exact VCS match',
        ...     pattern=re.compile(
        ...         rf'''
        ...         ^(?P<scheme>ssh)?
        ...         ((?P<user>\w+)@)?
        ...         (?P<hostname>(svn.project.org)+):
        ...         (?P<path>(\w[^:]+))
        ...         ''',
        ...         re.VERBOSE,
        ...     ),
        ...     is_explicit=True,
        ...     defaults={
        ...         'hostname': 'svn.project.org'
        ...     }
        ... )

        >>> SvnURL.rule_map.register(projectRule)

        >>> SvnURL.is_valid(
        ...     url='svn@svn.project.org:project-central/image', is_explicit=True
        ... )
        True

        >>> SvnURL(url='svn@svn.project.org:project-central/image').rule
        'project-rule'

        This is just us cleaning up:

        >>> SvnURL.rule_map.unregister('project-rule')

        >>> SvnURL(url='svn@svn.project.org:project-central/project-rule').rule
        'core-svn-scp'
        """
        return super().is_valid(url=url, is_explicit=is_explicit)

    def to_url(self) -> str:
        """Return a ``svn(1)``-compatible URL. Can be used with ``svn checkout``.

        Examples
        --------
        SSH style URL:

        >>> svn_url = SvnURL(url='svn@svn.project.org:project-central/browser')

        >>> svn_url.path = 'project-central/gfx'

        >>> svn_url.to_url()
        'svn@svn.project.org:project-central/gfx'

        HTTPs URL:

        >>> svn_url = SvnURL(url='https://svn.project.org/project-central/memory')

        >>> svn_url.path = 'project-central/image'

        >>> svn_url.to_url()
        'https://svn.project.org/project-central/image'

        Switch them to svnlab:

        >>> svn_url.hostname = 'localhost'
        >>> svn_url.scheme = 'http'

        >>> svn_url.to_url()
        'http://localhost/project-central/image'

        Pip style URL, thanks to this class implementing :class:`SvnPipURL`:

        >>> svn_url = SvnURL(url='svn+ssh://svn@svn.project.org/project-central/image')

        >>> svn_url.hostname = 'localhost'

        >>> svn_url.to_url()
        'svn+ssh://svn@localhost/project-central/image'

        >>> svn_url.user = None

        >>> svn_url.to_url()
        'svn+ssh://localhost/project-central/image'

        See Also
        --------
        :meth:`SvnBaseURL.to_url`, :meth:`SvnPipURL.to_url`
        """
        return super().to_url()



"""Detect, parse, and validate hg (Mercurial) URLs.

- Detect: :meth:`HgURL.is_valid()`
- Parse: :class:`HgURL`

  compare to :class:`urllib.parse.ParseResult`

  - Output ``hg(1)`` URL: :meth:`HgURL.to_url()`
- Extendable via :class:`~libvcs.url.base.RuleMap`,
  :class:`~libvcs.url.base.Rule`

.. Note::

   Mercurial URL support is experimental. If you rely on these URL shapes in
   production, please verify them against your hosts and report gaps in the
   `project tracker <https://github.com/vcs-python/libvcs>`_.
"""

from __future__ import annotations

import dataclasses
import re

from libvcs._internal.dataclasses import SkipDefaultFieldsReprMixin
from libvcs.url.git import RE_SUFFIX

from .base import Rule, RuleMap, URLProtocol
from .constants import RE_PIP_REV, RE_SCP, RE_USER

RE_PATH = r"""
    (?P<hostname>([^/:]+))
    (:(?P<port>\d{1,5}))?
    (?P<separator>[:,/])?
    (?P<path>
      /?(\w[^:.@]*)
    )?
"""


RE_SCHEME = r"""
    (?P<scheme>
      (
        http|https|ssh
      )
    )
"""

DEFAULT_RULES: list[Rule] = [
    Rule(
        label="core-hg",
        description="Vanilla hg pattern",
        pattern=re.compile(
            rf"""
        ^{RE_SCHEME}
        ://
        {RE_USER}
        {RE_PATH}
        {RE_SUFFIX}?
        {RE_PIP_REV}?
        """,
            re.VERBOSE,
        ),
    ),
    Rule(
        label="core-hg-scp",
        description="Vanilla scp(1) / ssh(1) type URL",
        pattern=re.compile(
            rf"""
        ^(?P<scheme>ssh)?
        {RE_USER}
        {RE_SCP}
        {RE_SUFFIX}?
        """,
            re.VERBOSE,
        ),
        defaults={"username": "hg"},
    ),
    # SCP-style URLs, e.g. hg@
]
"""Core regular expressions. These are patterns understood by ``hg(1)``"""


#
# Third-party URLs, e.g. pip, etc.
#
RE_PIP_SCHEME = r"""
    (?P<scheme>
      (
        hg\+ssh|
        hg\+https|
        hg\+http|
        hg\+file
      )
    )
"""

PIP_DEFAULT_RULES: list[Rule] = [
    Rule(
        label="pip-url",
        description="pip-style hg URL",
        pattern=re.compile(
            rf"""
        ^{RE_PIP_SCHEME}
        ://
        {RE_USER}
        {RE_PATH}
        {RE_SUFFIX}?
        {RE_PIP_REV}?
        """,
            re.VERBOSE,
        ),
        is_explicit=True,
    ),
    # file://, RTC 8089, File:// https://datatracker.ietf.org/doc/html/rfc8089
    Rule(
        label="pip-file-url",
        description="pip-style hg+file:// URL",
        pattern=re.compile(
            r"""
        (?P<scheme>hg\+file)://
        (?P<path>.*)
        """,
            re.VERBOSE,
        ),
        is_explicit=True,
    ),
]
"""pip-style hg URLs.

Examples of PIP-style hg URLs (via pip.pypa.io)::

    MyProject @ hg+http://hg.myproject.org/MyProject
    MyProject @ hg+https://hg.myproject.org/MyProject
    MyProject @ hg+ssh://hg.myproject.org/MyProject
    MyProject @ hg+file:///home/user/projects/MyProject

Refs (via pip.pypa.io)::

    MyProject @ hg+http://hg.example.com/MyProject@da39a3ee5e6b
    MyProject @ hg+http://hg.example.com/MyProject@2019
    MyProject @ hg+http://hg.example.com/MyProject@v1.0
    MyProject @ hg+http://hg.example.com/MyProject@special_feature

Notes
-----
- https://pip.pypa.io/en/stable/topics/vcs-support/
"""


@dataclasses.dataclass(repr=False)
class HgBaseURL(
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """Mercurial repository location. Parses URLs on initialization.

    Examples
    --------
    >>> HgBaseURL(url='https://hg.mozilla.org/mozilla-central/')
    HgBaseURL(url=https://hg.mozilla.org/mozilla-central/,
            scheme=https,
            hostname=hg.mozilla.org,
            path=mozilla-central/,
            rule=core-hg)

    >>> myrepo = HgURL(url='https://hg.mozilla.org/mozilla-central/')

    >>> myrepo.hostname
    'hg.mozilla.org'

    >>> myrepo.path
    'mozilla-central/'

    >>> HgBaseURL.is_valid(url='ssh://username@machinename/path/to/repo')
    True

    >>> HgBaseURL(url='ssh://username@machinename/path/to/repo')
    HgBaseURL(url=ssh://username@machinename/path/to/repo,
            scheme=ssh,
            user=username,
            hostname=machinename,
            path=path/to/repo,
            rule=core-hg)

    - Compatibility checking: :meth:`HgURL.is_valid()`
    - URLs compatible with ``hg(1)``: :meth:`HgURL.to_url()`
    """

    url: str
    scheme: str | None = None
    user: str | None = None
    hostname: str = dataclasses.field(default="")
    port: int | None = None
    separator: str = dataclasses.field(default="/")
    path: str = dataclasses.field(default="")

    # Decoration
    suffix: str | None = None

    #
    # commit-ish: tag, branch, ref, revision
    #
    ref: str | None = None

    rule: str | None = None
    # name of the :class:`Rule`

    rule_map = RuleMap(_rule_map={m.label: m for m in DEFAULT_RULES})

    def __post_init__(self) -> None:
        """Initialize :class:`~libvcs.url.hg.HgBaseURL` params into attributes."""
        url = self.url
        for rule in self.rule_map.values():
            match = re.match(rule.pattern, url)
            if match is None:
                continue
            groups = match.groupdict()
            self.rule = rule.label
            for k, v in groups.items():
                setattr(self, k, v)

            for k in rule.defaults:
                if getattr(self, k, None) is None:
                    setattr(self, k, rule.defaults[k])

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        """Whether URL is compatible with VCS or not.

        Examples
        --------
        >>> HgBaseURL.is_valid(
        ...     url='https://hg.mozilla.org/mozilla-central'
        ... )
        True

        >>> HgBaseURL.is_valid(url='ssh://hg@hg.python.org/cpython')
        True

        >>> HgBaseURL.is_valid(url='notaurl')
        False
        """
        if is_explicit is not None:
            return any(
                re.search(rule.pattern, url)
                for rule in cls.rule_map.values()
                if rule.is_explicit == is_explicit
            )
        return any(re.search(rule.pattern, url) for rule in cls.rule_map.values())

    def to_url(self) -> str:
        """Return a ``hg(1)``-compatible URL. Can be used with ``hg clone``.

        Examples
        --------
        >>> hg_url = HgBaseURL(url='https://hg.mozilla.org/mozilla-central')

        >>> hg_url
        HgBaseURL(url=https://hg.mozilla.org/mozilla-central,
                scheme=https,
                hostname=hg.mozilla.org,
                path=mozilla-central,
                rule=core-hg)

        Switch repo libvcs -> vcspull:

        >>> hg_url.path = 'mobile-browser'

        >>> hg_url.to_url()
        'https://hg.mozilla.org/mobile-browser'

        Switch them to localhost:

        >>> hg_url.hostname = 'localhost'
        >>> hg_url.scheme = 'http'

        >>> hg_url.to_url()
        'http://localhost/mobile-browser'

        Another example, `hugin <http://hugin.hg.sourceforge.net>`_:

        >>> hugin = HgBaseURL(
        ...     url="http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin")

        >>> hugin
        HgBaseURL(url=http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin,
                scheme=http,
                hostname=hugin.hg.sourceforge.net,
                port=8000,
                path=hgroot/hugin/hugin,
                rule=core-hg)

        >>> hugin.to_url()
        'http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin'

        SSH URL with a username, `graphicsmagic <http://graphicsmagick.org/Hg.html>`_:

        >>> graphicsmagick = HgBaseURL(
        ...     url="ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick"
        ... )

        >>> graphicsmagick
        HgBaseURL(url=ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick,
                scheme=ssh,
                user=yourid,
                hostname=hg.GraphicsMagick.org,
                path=/hg/GraphicsMagick,
                rule=core-hg)

        >>> graphicsmagick.to_url()
        'ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick'

        Switch the username:

        >>> graphicsmagick.user = 'lucas'

        >>> graphicsmagick.to_url()
        'ssh://lucas@hg.GraphicsMagick.org//hg/GraphicsMagick'

        """
        if self.scheme is not None:
            parts = [self.scheme, "://"]

            if self.user is not None:
                parts.append(f"{self.user}@")
            parts.append(self.hostname)
        else:
            parts = [self.user or "hg", "@", self.hostname]

        if self.port is not None:
            parts.extend([":", f"{self.port}"])

        parts.extend([self.separator, self.path])

        return "".join(part for part in parts if isinstance(part, str))


@dataclasses.dataclass(repr=False)
class HgPipURL(
    HgBaseURL,
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """Supports pip hg URLs."""

    # commit-ish (rev): tag, branch, ref
    rev: str | None = None

    rule_map = RuleMap(_rule_map={m.label: m for m in PIP_DEFAULT_RULES})

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        """Whether URL is compatible with VCS or not.

        Examples
        --------
        >>> HgPipURL.is_valid(
        ...     url='hg+https://hg.mozilla.org/mozilla-central'
        ... )
        True

        >>> HgPipURL.is_valid(url='hg+ssh://hg@hg.python.org:cpython')
        True

        >>> HgPipURL.is_valid(url='notaurl')
        False
        """
        return super().is_valid(url=url, is_explicit=is_explicit)

    def to_url(self) -> str:
        """Return a ``hg(1)``-compatible URL. Can be used with ``hg clone``.

        Examples
        --------
        >>> hg_url = HgPipURL(url='hg+https://hg.mozilla.org/mozilla-central')

        >>> hg_url
        HgPipURL(url=hg+https://hg.mozilla.org/mozilla-central,
                scheme=hg+https,
                hostname=hg.mozilla.org,
                path=mozilla-central,
                rule=pip-url)

        Switch repo mozilla-central -> mobile-browser:

        >>> hg_url.path = 'mobile-browser'

        >>> hg_url.to_url()
        'hg+https://hg.mozilla.org/mobile-browser'

        Switch them to localhost:

        >>> hg_url.hostname = 'localhost'
        >>> hg_url.scheme = 'http'

        >>> hg_url.to_url()
        'http://localhost/mobile-browser'

        """
        parts = [self.scheme or "ssh", "://"]
        if self.user:
            parts.extend([self.user, "@"])

        parts.append(self.hostname)

        if self.port is not None:
            parts.extend([":", f"{self.port}"])

        parts.extend([self.separator, self.path])

        return "".join(part for part in parts if isinstance(part, str))


@dataclasses.dataclass(repr=False)
class HgURL(
    HgPipURL,
    HgBaseURL,
    URLProtocol,
    SkipDefaultFieldsReprMixin,
):
    """Batteries included URL Parser. Supports hg(1) and pip URLs.

    **Ancestors (MRO)**
    This URL parser inherits methods and attributes from the following parsers:

    - :class:`HgPipURL`

      - :meth:`HgPipURL.to_url`
    - :class:`HgBaseURL`

      - :meth:`HgBaseURL.to_url`
    """

    rule_map = RuleMap(
        _rule_map={m.label: m for m in [*DEFAULT_RULES, *PIP_DEFAULT_RULES]},
    )

    @classmethod
    def is_valid(cls, url: str, is_explicit: bool | None = None) -> bool:
        r"""Whether URL is compatible included Hg URL rule_map or not.

        Examples
        --------
        **Will** match normal ``hg(1)`` URLs, use :meth:`HgURL.is_valid` for that.

        >>> HgURL.is_valid(url='https://hg.mozilla.org/mozilla-central/mozilla-central')
        True

        >>> HgURL.is_valid(url='hg@hg.mozilla.org:MyProject/project')
        True

        Pip-style URLs:

        >>> HgURL.is_valid(url='hg+https://hg.mozilla.org/mozilla-central/project')
        True

        >>> HgURL.is_valid(url='hg+ssh://hg@hg.mozilla.org:MyProject/project')
        True

        >>> HgURL.is_valid(url='notaurl')
        False

        **Explicit VCS detection**

        Pip-style URLs are prefixed with the VCS name in front, so its rule_map can
        unambiguously narrow the type of VCS:

        >>> HgURL.is_valid(
        ...     url='hg+ssh://hg@hg.mozilla.org:mozilla-central/image', is_explicit=True
        ... )
        True

        Below, while it's hg.mozilla.org, that doesn't necessarily mean that the URL
        itself is conclusively a `hg` URL (e.g. the pattern is too broad):

        >>> HgURL.is_valid(
        ...     url='hg@hg.mozilla.org:mozilla-central/image', is_explicit=True
        ... )
        False

        You could create a Mozilla rule that consider hg.mozilla.org hostnames to be
        exclusively hg:

        >>> MozillaRule = Rule(
        ...     # Since hg.mozilla.org exclusively serves hg repos, make explicit
        ...     label='mozilla-rule',
        ...     description='Matches hg.mozilla.org https URLs, exact VCS match',
        ...     pattern=re.compile(
        ...         rf'''
        ...         ^(?P<scheme>ssh)?
        ...         ((?P<user>\w+)@)?
        ...         (?P<hostname>(hg.mozilla.org)+):
        ...         (?P<path>(\w[^:]+))
        ...         {RE_SUFFIX}?
        ...         ''',
        ...         re.VERBOSE,
        ...     ),
        ...     is_explicit=True,
        ...     defaults={
        ...         'hostname': 'hg.mozilla.org'
        ...     }
        ... )

        >>> HgURL.rule_map.register(MozillaRule)

        >>> HgURL.is_valid(
        ...     url='hg@hg.mozilla.org:mozilla-central/image', is_explicit=True
        ... )
        True

        >>> HgURL(url='hg@hg.mozilla.org:mozilla-central/image').rule
        'mozilla-rule'

        This is just us cleaning up:

        >>> HgURL.rule_map.unregister('mozilla-rule')

        >>> HgURL(url='hg@hg.mozilla.org:mozilla-central/mozilla-rule').rule
        'core-hg-scp'
        """
        return super().is_valid(url=url, is_explicit=is_explicit)

    def to_url(self) -> str:
        """Return a ``hg(1)``-compatible URL. Can be used with ``hg clone``.

        Examples
        --------
        SSH style URL:

        >>> hg_url = HgURL(url='hg@hg.mozilla.org:mozilla-central/browser')

        >>> hg_url.path = 'mozilla-central/gfx'

        >>> hg_url.to_url()
        'ssh://hg@hg.mozilla.org:mozilla-central/gfx'

        HTTPs URL:

        >>> hg_url = HgURL(url='https://hg.mozilla.org/mozilla-central/memory')

        >>> hg_url.path = 'mozilla-central/image'

        >>> hg_url.to_url()
        'https://hg.mozilla.org/mozilla-central/image'

        Switch them to hglab:

        >>> hg_url.hostname = 'localhost'
        >>> hg_url.scheme = 'http'

        >>> hg_url.to_url()
        'http://localhost/mozilla-central/image'

        Pip style URL, thanks to this class implementing :class:`HgPipURL`:

        >>> hg_url = HgURL(url='hg+ssh://hg@hg.mozilla.org/mozilla-central/image')

        >>> hg_url.hostname = 'localhost'

        >>> hg_url.to_url()
        'hg+ssh://hg@localhost/mozilla-central/image'

        >>> hg_url.user = None

        >>> hg_url.to_url()
        'hg+ssh://localhost/mozilla-central/image'

        See Also
        --------
        :meth:`HgBaseURL.to_url`, :meth:`HgPipURL.to_url`
        """
        return super().to_url()



"""Tool to manage a local hg (Mercurial) working copy from a repository.

.. todo::

   The following is from pypa/pip (MIT license):

   - :meth:`~libvcs.sync.base.BaseSync.from_pip_url`
   - :meth:`~libvcs.sync.hg.HgSync.get_revision`
"""  # E5

from __future__ import annotations

import logging
import pathlib
import typing as t

from libvcs import exc
from libvcs._internal.types import StrPath
from libvcs.cmd.hg import Hg

from .base import BaseSync, SyncResult

logger = logging.getLogger(__name__)


class HgSync(BaseSync):
    """Tool to manage a local hg (Mercurial) repository cloned from a remote one."""

    bin_name = "hg"
    schemes = ("hg", "hg+http", "hg+https", "hg+file")
    cmd: Hg

    def __init__(
        self,
        *,
        url: str,
        path: StrPath,
        **kwargs: t.Any,
    ) -> None:
        """Local Mercurial repository.

        Parameters
        ----------
        url : str
            Mercurial repository URL.
        """
        super().__init__(url=url, path=path, **kwargs)

        self.cmd = Hg(path=path, progress_callback=self.progress_callback)

    def obtain(self, *args: t.Any, **kwargs: t.Any) -> None:
        """Clone and update a Mercurial repository to this location."""
        self.cmd.clone(
            no_update=True,
            quiet=True,
            url=self.url,
        )
        self.cmd.update(
            quiet=True,
            check_returncode=True,
        )

    def get_revision(self) -> str:
        """Get latest revision of this mercurial repository."""
        return self.run(["parents", "--template={rev}"])

    def update_repo(self, *args: t.Any, **kwargs: t.Any) -> SyncResult:
        """Pull changes from remote Mercurial repository into this one.

        Returns
        -------
        SyncResult
            Result of the sync operation, with any errors recorded.
        """
        result = SyncResult()
        if not pathlib.Path(self.path / ".hg").exists():
            try:
                self.obtain()
            except exc.CommandError as e:
                self.log.exception("Failed to obtain repository")
                result.add_error("obtain", str(e), exception=e)
                return result
            return self.update_repo()
        else:
            try:
                self.cmd.update()
                self.cmd.pull(update=True)
            except exc.CommandError as e:
                result.add_error("pull", str(e), exception=e)
        return result



"""Misc. legacy helpers :mod:`subprocess` and finding VCS binaries.

:class:`libvcs._internal.run.run` will be deprecated by
:mod:`libvcs._internal.subprocess`.

Note
----
This is an internal API not covered by versioning policy.
"""

from __future__ import annotations

import contextlib
import datetime
import logging
import os
import selectors
import subprocess
import sys
import time
import typing as t
from collections.abc import Iterable, Mapping, MutableMapping, Sequence

from libvcs import exc
from libvcs._internal.types import StrOrBytesPath

logger = logging.getLogger(__name__)

console_encoding = sys.stdout.encoding


def console_to_str(s: bytes) -> str:
    """From pypa/pip project, pip.backwardwardcompat. License MIT."""
    try:
        return s.decode(console_encoding)
    except UnicodeDecodeError:
        return s.decode("utf_8", errors="backslashreplace")
    except AttributeError:  # for tests, #13
        return str(s)


if t.TYPE_CHECKING:
    _LoggerAdapter = logging.LoggerAdapter[logging.Logger]
else:
    _LoggerAdapter = logging.LoggerAdapter


class CmdLoggingAdapter(_LoggerAdapter):
    """Adapter for additional command-related data to :py:mod:`logging`.

    Extends :py:class:`logging.LoggerAdapter`'s functionality.

    Mixes in additional context via :py:meth:`logging.LoggerAdapter.process()` for
    :class:`logging.Formatter` when emitting log entries.

    Parameters
    ----------
    bin_name : str
        name of the command or vcs tool being wrapped, e.g. 'git'
    keyword : str
        directory basename, name of repo, hint, etc. e.g. 'django'
    """

    def __init__(
        self,
        bin_name: str,
        keyword: str,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        #: bin_name
        self.bin_name = bin_name
        #: directory basename, name of repository, hint, etc.
        self.keyword = keyword

        logging.LoggerAdapter.__init__(self, *args, **kwargs)

    def process(
        self,
        msg: str,
        kwargs: MutableMapping[str, t.Any],
    ) -> tuple[t.Any, MutableMapping[str, t.Any]]:
        """Add additional context information for loggers."""
        prefixed_dict = {}
        prefixed_dict["bin_name"] = self.bin_name
        prefixed_dict["keyword"] = self.keyword

        kwargs["extra"] = prefixed_dict

        return msg, kwargs


class ProgressCallbackProtocol(t.Protocol):
    """Callback to report subprocess communication."""

    def __call__(self, output: str, timestamp: datetime.datetime) -> None:
        """Process progress for subprocess communication."""
        ...


if sys.platform == "win32":
    _ENV: t.TypeAlias = Mapping[str, str]
else:
    _ENV: t.TypeAlias = Mapping[bytes, StrOrBytesPath] | Mapping[str, StrOrBytesPath]

_CMD: t.TypeAlias = StrOrBytesPath | Sequence[StrOrBytesPath]
_FILE: t.TypeAlias = int | t.IO[t.Any] | None


def _normalize_command_args(args: _CMD) -> list[StrOrBytesPath]:
    """Return subprocess arguments without splitting scalar strings or bytes."""
    if isinstance(args, (str, bytes, os.PathLike)):
        return [os.fspath(args)]

    return [os.fspath(arg) for arg in args]


def _stringify_command(args: _CMD) -> str | list[str]:
    """Return a human-readable command for CommandError."""
    if isinstance(args, (str, bytes, os.PathLike)):
        return os.fsdecode(args)

    return [os.fsdecode(arg) for arg in args]


def run(
    args: _CMD,
    bufsize: int = -1,
    executable: StrOrBytesPath | None = None,
    stdin: _FILE | None = None,
    stdout: _FILE | None = None,
    stderr: _FILE | None = None,
    preexec_fn: t.Callable[[], t.Any] | None = None,
    close_fds: bool = True,
    shell: bool = False,
    cwd: StrOrBytesPath | None = None,
    env: _ENV | None = None,
    startupinfo: t.Any | None = None,
    creationflags: int = 0,
    restore_signals: bool = True,
    start_new_session: bool = False,
    pass_fds: t.Any = (),
    *,
    encoding: str | None = None,
    errors: str | None = None,
    user: str | int | None = None,
    group: str | int | None = None,
    extra_groups: Iterable[str | int] | None = None,
    umask: int = -1,
    log_in_real_time: bool = False,
    check_returncode: bool = True,
    trim: bool = False,
    callback: ProgressCallbackProtocol | None = None,
    timeout: float | None = None,
) -> str:
    """Run a command.

    Run 'args' in a shell and return the combined contents of stdout and
    stderr (Blocking). Throws an exception if the command exits non-zero.

    Output is returned verbatim by default, so whitespace-significant output
    round-trips exactly -- for example a ``git diff`` destined for
    ``git apply``, which requires the trailing newline, or a ``git cat-file
    blob`` whose contents must match byte-for-byte. Pass ``trim=True`` to strip
    trailing whitespace from the whole output for convenient "bare value" reads
    where a trailing newline is just noise (e.g. ``rev-parse HEAD``).

    Keyword arguments are passthrough to :class:`subprocess.Popen`.

    Parameters
    ----------
    args : list or str, or single str, if shell=True
       the command to run

    shell : bool
        boolean indicating whether we are using advanced shell
        features. Use only when absolutely necessary, since this allows a lot
        more freedom which could be exploited by malicious code. See the
        warning here: http://docs.python.org/library/subprocess.html#popen-constructor

    cwd : str
        dir command is run from. Defaults to ``path``.

    log_in_real_time : bool
        boolean indicating whether to read stdout from the
        subprocess in real time instead of when the process finishes.

    check_returncode : bool
        Indicate whether a ``libvcs.exc.CommandError`` should be raised if return
        code is different from 0.

    trim : bool
        When False (the default), return the output verbatim, including any
        trailing newline, so whitespace-significant output (diffs, blob
        contents) round-trips exactly. When True, strip trailing whitespace
        from the whole output for the convenient "bare value" reads callers
        expect (e.g. ``rev-parse HEAD``). The same policy applies to stderr
        captured for a failed command's error output.

    callback : ProgressCallbackProtocol
        callback to return output as a command executes, accepts a function signature
        of ``(output, timestamp)``. Example usage::

            def progress_cb(output, timestamp):
                sys.stdout.write(output)
                sys.stdout.flush()
            run(['git', 'pull'], callback=progress_cb)

    timeout : float, optional
        Seconds to wait before terminating the subprocess. When the deadline is
        exceeded the process is sent ``SIGTERM`` (then ``SIGKILL`` after a short
        grace period) and :class:`libvcs.exc.CommandTimeoutError` is raised with
        any output collected so far. ``None`` (default) disables the deadline and
        preserves the legacy behaviour of blocking until the process exits.

    Upcoming changes
    ----------------
    When minimum python >= 3.10, pipesize: int = -1 will be added after umask.
    """
    normalized_args: _CMD
    if shell:
        normalized_args = os.fspath(args) if isinstance(args, os.PathLike) else args
    else:
        normalized_args = _normalize_command_args(args)

    proc = subprocess.Popen(
        normalized_args,
        bufsize=bufsize,
        executable=executable,
        stdin=stdin,
        stdout=stdout or subprocess.PIPE,
        stderr=stderr or subprocess.PIPE,
        preexec_fn=preexec_fn,
        close_fds=close_fds,
        shell=shell,
        cwd=cwd,
        env=env,
        startupinfo=startupinfo,
        creationflags=creationflags,
        restore_signals=restore_signals,
        start_new_session=start_new_session,
        pass_fds=pass_fds,
        text=False,  # Keep in bytes mode to preserve \r properly
        encoding=encoding,
        errors=errors,
        user=user,
        group=group,
        extra_groups=extra_groups,
        umask=umask,
    )

    all_output: str = ""
    code = None
    line = None
    if log_in_real_time and callback is None:

        def progress_cb(output: t.AnyStr, timestamp: datetime.datetime) -> None:
            sys.stdout.write(str(output))
            sys.stdout.flush()

        callback = progress_cb

    # Note: When git detects that stderr is not a TTY (e.g., when piped),
    # it outputs progress with newlines instead of carriage returns.
    # This causes each progress update to appear on a new line.
    # To get proper single-line progress updates, git would need to be
    # connected to a pseudo-TTY, which would require significant changes
    # to how subprocess execution is handled.

    timeout_stdout: bytes | None = None
    timeout_stderr: bytes | None = None
    if timeout is None:
        while code is None:
            code = proc.poll()

            if callback and callable(callback) and proc.stderr is not None:
                line = console_to_str(proc.stderr.read(128))
                if line:
                    callback(output=line, timestamp=datetime.datetime.now())
    else:
        code, timeout_stdout, timeout_stderr = _wait_with_deadline(
            proc,
            deadline=time.monotonic() + timeout,
            timeout=timeout,
            callback=callback,
            cmd=_stringify_command(normalized_args),
        )
    if callback and callable(callback):
        callback(output="\r", timestamp=datetime.datetime.now())

    if proc.stdout is not None:
        raw_stdout: bytes = (
            timeout_stdout
            if timeout_stdout is not None
            else b"".join(proc.stdout.readlines())
        )
        all_output = console_to_str(raw_stdout)
        if trim:
            all_output = all_output.rstrip()
    else:
        all_output = ""
    if code and proc.stderr is not None:
        raw_stderr: bytes = (
            timeout_stderr
            if timeout_stderr is not None
            else b"".join(proc.stderr.readlines())
        )
        all_output = console_to_str(raw_stderr)
        if trim:
            all_output = all_output.rstrip()
    output = "".join(all_output)
    if code != 0 and check_returncode:
        raise exc.CommandError(
            output=output,
            returncode=code,
            cmd=_stringify_command(normalized_args),
        )
    return output


#: Grace period after ``terminate()`` before escalating to ``kill()``.
_TIMEOUT_KILL_GRACE_SECONDS = 0.5

#: Upper bound on the ``selectors.select()`` wait inside the deadline loop.
_TIMEOUT_POLL_INTERVAL_SECONDS = 0.1


def _wait_with_deadline(
    proc: subprocess.Popen[bytes],
    *,
    deadline: float,
    timeout: float,
    callback: ProgressCallbackProtocol | None,
    cmd: str | list[str],
) -> tuple[int, bytes | None, bytes | None]:
    """Wait for ``proc`` to exit, enforcing a wall-clock deadline.

    Drains both ``stdout`` and ``stderr`` concurrently so a child that fills
    either kernel pipe buffer (~64 KiB on Linux) cannot deadlock waiting for
    libvcs to read its output. Uses :mod:`selectors` to unblock when either
    stream is readable, when the child exits, or when the per-iteration poll
    interval expires -- whichever comes first.

    When the deadline is exceeded the subprocess is reaped and
    :class:`libvcs.exc.CommandTimeoutError` is raised with the bytes captured
    before the timeout. Otherwise the captured stdout/stderr are returned to
    the caller alongside the exit code so they can be reused without re-
    reading the now-drained pipes.

    Returns
    -------
    tuple[int, bytes | None, bytes | None]
        ``(returncode, stdout_bytes, stderr_bytes)``. A byte buffer is
        ``None`` when the corresponding pipe could not be put into non-
        blocking mode (Windows pipes, unusual fd types) -- in that case the
        caller should fall back to reading the pipe directly.

    Notes
    -----
    The progress ``callback`` is invoked for ``stderr`` chunks only,
    matching the legacy non-timeout codepath. ``stdout`` is drained into
    the returned buffer to prevent the child from blocking on a full pipe,
    but its chunks are not forwarded to the callback in real time. Callers
    that want streaming ``stdout`` should redirect it themselves
    (e.g. ``stdout=`` to a file) rather than relying on ``callback``.
    """
    sel = selectors.DefaultSelector()
    buffers: dict[t.IO[bytes], list[bytes]] = {}
    registered: set[t.IO[bytes]] = set()
    fds_to_restore: list[int] = []

    for stream in (proc.stdout, proc.stderr):
        if stream is None:
            continue
        # Non-blocking reads so the selector loop never stalls on a short read
        # and a chatty child cannot fill a pipe buffer and wedge itself.
        try:
            fd = stream.fileno()
            os.set_blocking(fd, False)
        except (OSError, ValueError):
            continue
        sel.register(stream, selectors.EVENT_READ)
        buffers[stream] = []
        registered.add(stream)
        fds_to_restore.append(fd)

    code: int | None = None
    try:
        while True:
            code = proc.poll()
            if code is not None:
                # Final drain: data written between the last ``select()`` wake
                # and the child's exit would otherwise be lost on the early
                # return. Only drain streams we put into non-blocking mode.
                for stream in list(registered):
                    trailing = _drain_stream(stream)
                    if trailing:
                        buffers[stream].append(trailing)
                        if (
                            stream is proc.stderr
                            and callback is not None
                            and callable(callback)
                        ):
                            callback(
                                output=console_to_str(trailing),
                                timestamp=datetime.datetime.now(),
                            )
                break

            remaining = deadline - time.monotonic()
            if remaining <= 0:
                # ``vcs_exit_code`` deliberately omitted here: ``proc.returncode``
                # is still ``None`` because the child has not been signalled yet,
                # and CLAUDE.md treats ``vcs_exit_code`` as a scalar ``int``.
                logger.warning(
                    "subprocess deadline exceeded after %.3gs",
                    timeout,
                    extra={"vcs_cmd": _format_cmd_for_log(cmd)},
                )
                _terminate_process(proc, cmd)
                for stream in list(registered):
                    trailing = _drain_stream(stream)
                    if trailing:
                        buffers[stream].append(trailing)
                stdout_data = _join_buffer(buffers, proc.stdout)
                stderr_data = _join_buffer(buffers, proc.stderr)
                message = console_to_str((stdout_data or b"") + (stderr_data or b""))
                raise exc.CommandTimeoutError(
                    output=message,
                    returncode=proc.returncode,
                    cmd=cmd,
                    timeout=timeout,
                )

            wait = min(_TIMEOUT_POLL_INTERVAL_SECONDS, remaining)
            if not registered:
                # No streams to select on (e.g. ``os.set_blocking`` failed on
                # Windows pipes). Yield the CPU explicitly instead of busy-
                # looping until the deadline or process exit.
                time.sleep(wait)
                continue

            events = sel.select(timeout=wait)
            if not events:
                continue

            for key, _mask in events:
                stream = t.cast("t.IO[bytes]", key.fileobj)
                try:
                    chunk = stream.read(128)
                except (BlockingIOError, OSError):
                    chunk = b""
                if not chunk:
                    # EOF from the child closing the pipe. Stop selecting on
                    # it so the loop doesn't spin on an always-readable fd.
                    sel.unregister(stream)
                    registered.discard(stream)
                    continue
                buffers[stream].append(chunk)
                if (
                    stream is proc.stderr
                    and callback is not None
                    and callable(callback)
                ):
                    callback(
                        output=console_to_str(chunk),
                        timestamp=datetime.datetime.now(),
                    )
    finally:
        # Restore blocking mode so any subsequent read by the caller behaves
        # as expected; ignore failures (fd already closed, Windows pipe).
        for fd in fds_to_restore:
            with contextlib.suppress(OSError, ValueError):
                os.set_blocking(fd, True)
        sel.close()

    return code, _join_buffer(buffers, proc.stdout), _join_buffer(buffers, proc.stderr)


def _join_buffer(
    buffers: dict[t.IO[bytes], list[bytes]],
    stream: t.IO[bytes] | None,
) -> bytes | None:
    """Concatenate captured chunks for ``stream``, or ``None`` if not tracked."""
    if stream is None or stream not in buffers:
        return None
    return b"".join(buffers[stream])


def _terminate_process(proc: subprocess.Popen[bytes], cmd: str | list[str]) -> None:
    """Terminate ``proc`` gracefully, falling back to ``kill`` on the grace."""
    if proc.poll() is not None:
        return
    try:
        proc.terminate()
    except (OSError, ProcessLookupError):
        return
    try:
        proc.wait(timeout=_TIMEOUT_KILL_GRACE_SECONDS)
    except subprocess.TimeoutExpired:
        logger.debug(
            "subprocess sigkill escalated after sigterm grace expired",
            extra={"vcs_cmd": _format_cmd_for_log(cmd)},
        )
        with contextlib.suppress(OSError, ProcessLookupError):
            proc.kill()
        # If the child is still unreachable after SIGKILL, bail rather than
        # block forever -- we've already signalled the user-facing timeout.
        with contextlib.suppress(subprocess.TimeoutExpired):
            proc.wait(timeout=_TIMEOUT_KILL_GRACE_SECONDS)
        if proc.poll() is None:
            logger.warning(
                "subprocess sigkill did not reap; child may be leaked",
                extra={"vcs_cmd": _format_cmd_for_log(cmd)},
            )


def _format_cmd_for_log(cmd: str | list[str]) -> str:
    """Render ``cmd`` as a flat string for the ``vcs_cmd`` log extra."""
    if isinstance(cmd, list):
        return " ".join(cmd)
    return cmd


def _drain_stream(stream: t.IO[bytes] | None) -> bytes:
    """Best-effort read of any remaining bytes from a subprocess pipe."""
    if stream is None:
        return b""
    try:
        data = stream.read() or b""
    except (BlockingIOError, OSError, ValueError):
        data = b""
    return data



"""Shortcuts for creating repos.

Note
----
This is an internal API not covered by versioning policy.
"""

from __future__ import annotations

import typing as t

from libvcs import GitSync, HgSync, SvnSync, exc
from libvcs._internal.run import ProgressCallbackProtocol
from libvcs._internal.types import StrPath, VCSLiteral
from libvcs.exc import InvalidVCS
from libvcs.url import registry as url_tools


class VCSNoMatchFoundForUrl(exc.LibVCSException):
    def __init__(self, url: str, *args: object) -> None:
        return super().__init__(f"No VCS found for url: {url}")


class VCSMultipleMatchFoundForUrl(exc.LibVCSException):
    def __init__(self, url: str, *args: object) -> None:
        return super().__init__(f"Multiple VCS found for url: {url}")


class VCSNotSupported(exc.LibVCSException):
    def __init__(self, url: str, vcs: str, *args: object) -> None:
        return super().__init__(f"VCS '{vcs}' not supported, based on URL: {url}")


@t.overload
def create_project(
    *,
    url: str,
    path: StrPath,
    vcs: t.Literal["git"],
    progress_callback: ProgressCallbackProtocol | None = None,
    **kwargs: t.Any,
) -> GitSync: ...


@t.overload
def create_project(
    *,
    url: str,
    path: StrPath,
    vcs: t.Literal["svn"],
    progress_callback: ProgressCallbackProtocol | None = None,
    **kwargs: t.Any,
) -> SvnSync: ...


@t.overload
def create_project(
    *,
    url: str,
    path: StrPath,
    vcs: t.Literal["hg"],
    progress_callback: ProgressCallbackProtocol | None = ...,
    **kwargs: t.Any,
) -> HgSync: ...


@t.overload
def create_project(
    *,
    url: str,
    path: StrPath,
    vcs: None = None,
    progress_callback: ProgressCallbackProtocol | None = None,
    **kwargs: t.Any,
) -> GitSync | HgSync | SvnSync: ...


def create_project(
    *,
    url: str,
    path: StrPath,
    vcs: VCSLiteral | None = None,
    progress_callback: ProgressCallbackProtocol | None = None,
    **kwargs: t.Any,
) -> GitSync | HgSync | SvnSync:
    r"""Return an object representation of a VCS repository.

    Examples
    --------
    >>> from libvcs._internal.shortcuts import create_project
    >>> r = create_project(
    ...     url=f'file://{create_git_remote_repo()}',
    ...     vcs='git',
    ...     path=tmp_path
    ... )

    >>> isinstance(r, GitSync)
    True

    create_project can also guess VCS for certain URLs:

    >>> r = create_project(
    ...     # Note the git+ before the URL
    ...     url=f'git+file://{create_git_remote_repo()}',
    ...     path=tmp_path
    ... )

    >>> isinstance(r, GitSync)
    True

    It also supports unprefixed SSH-style Git URLs:

    >>> r = create_project(
    ...     url='git@github.com:tmux-python/tmuxp.git',
    ...     path=tmp_path
    ... )
    >>> isinstance(r, GitSync)
    True
    """
    if vcs is None:
        vcs_matches = url_tools.registry.match(url=url, is_explicit=True)

        if len(vcs_matches) == 0:
            raise VCSNoMatchFoundForUrl(url=url)
        if len(vcs_matches) > 1:
            raise VCSMultipleMatchFoundForUrl(url=url)

        assert vcs_matches[0].vcs is not None

        def is_vcs(val: t.Any) -> t.TypeGuard[VCSLiteral]:
            return isinstance(val, str) and val in {"git", "hg", "svn"}

        if is_vcs(vcs_matches[0].vcs):
            vcs = vcs_matches[0].vcs
        else:
            raise VCSNotSupported(url=url, vcs=vcs_matches[0].vcs)

    if vcs == "git":
        return GitSync(
            url=url,
            path=path,
            progress_callback=progress_callback,
            **kwargs,
        )
    if vcs == "hg":
        return HgSync(url=url, path=path, progress_callback=progress_callback, **kwargs)
    if vcs == "svn":
        return SvnSync(
            url=url,
            path=path,
            progress_callback=progress_callback,
            **kwargs,
        )
    msg = f"VCS {vcs} is not a valid VCS"
    raise InvalidVCS(msg)



#!/usr/bin/env python
"""Tool to manage a local SVN (Subversion) working copy from a repository.

.. todo::

    The follow are from saltstack/salt (Apache license):

    - :meth:`~libvcs.sync.svn.SvnSync.get_revision_file`

    The following are pypa/pip (MIT license):

    - :meth:`~libvcs.sync.svn.SvnSync.get_revision`
"""

from __future__ import annotations

import logging
import os
import pathlib
import re
import typing as t

from libvcs import exc
from libvcs._internal.types import StrPath
from libvcs.cmd.svn import Svn

from .base import BaseSync, SyncResult

logger = logging.getLogger(__name__)


class SvnUrlRevFormattingError(ValueError):
    """Raised when SVN Revision output is not in the expected format."""

    def __init__(self, data: str, *args: object) -> None:
        return super().__init__(f"Badly formatted data: {data!r}")


class SvnSync(BaseSync):
    """Tool to manage a local SVN (Subversion) working copy from a SVN repository."""

    bin_name = "svn"
    schemes = ("svn", "svn+ssh", "svn+http", "svn+https", "svn+svn")
    cmd: Svn

    def __init__(
        self,
        *,
        url: str,
        path: StrPath,
        **kwargs: t.Any,
    ) -> None:
        """Working copy of a SVN repository.

        Parameters
        ----------
        url : str
            URL in subversion repository

        username : str, optional
            username to use for checkout and update

        password : str, optional
            password to use for checkout and update

        svn_trust_cert : bool
            trust the Subversion server site certificate, default False
        """
        self.svn_trust_cert = kwargs.pop("svn_trust_cert", False)

        self.username = kwargs.get("username")
        self.password = kwargs.get("password")

        self.rev = kwargs.get("rev")

        super().__init__(url=url, path=path, **kwargs)

        self.cmd = Svn(path=path, progress_callback=self.progress_callback)

    def _user_pw_args(self) -> list[t.Any]:
        args = []
        for param_name in ["svn_username", "svn_password"]:
            if hasattr(self, param_name):
                args.extend(["--" + param_name[4:], getattr(self, param_name)])
        return args

    def obtain(self, quiet: bool | None = None, *args: t.Any, **kwargs: t.Any) -> None:
        """Check out a working copy from a SVN repository."""
        url, rev = self.url, self.rev

        if rev is not None:
            kwargs["revision"] = rev
        if self.svn_trust_cert:
            kwargs["trust_server_cert"] = True
        self.cmd.checkout(
            url=url,
            username=self.username,
            password=self.password,
            non_interactive=True,
            quiet=True,
            check_returncode=True,
            **kwargs,
        )

    def get_revision_file(self, location: str) -> int:
        """Return revision for a file."""
        current_rev = self.cmd.info(location)

        INI_RE = re.compile(r"^([^:]+):\s+(\S.*)$", re.MULTILINE)

        info_list = INI_RE.findall(current_rev)
        return int(dict(info_list)["Revision"])

    def get_revision(self, location: str | None = None) -> int:
        """Return maximum revision for all files under a given location."""
        if not location:
            location = self.url

        if pathlib.Path(location).exists() and not pathlib.Path(location).is_dir():
            return self.get_revision_file(location)

        # Note: taken from setuptools.command.egg_info
        revision = 0

        for base, dirs, _files in os.walk(location):
            if ".svn" not in dirs:
                dirs[:] = []
                continue  # no sense walking uncontrolled subdirs
            dirs.remove(".svn")
            entries_fn = pathlib.Path(base) / ".svn" / "entries"
            if not entries_fn.exists():
                # FIXME: should we warn?
                continue

            dirurl, localrev = SvnSync._get_svn_url_rev(base)

            if base == location:
                assert dirurl is not None
                base = dirurl + "/"  # save the root url
            elif not dirurl or not dirurl.startswith(base):
                dirs[:] = []
                continue  # not part of the same svn tree, skip it
            revision = max(revision, localrev)
        return revision

    def update_repo(
        self,
        dest: str | None = None,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> SyncResult:
        """Fetch changes from SVN repository to local working copy.

        Parameters
        ----------
        dest : str or None, optional
            Destination path override for the working copy.

        Returns
        -------
        SyncResult
            Result of the sync operation, with any errors recorded.
        """
        result = SyncResult()
        self.ensure_dir()
        if pathlib.Path(self.path / ".svn").exists():
            try:
                self.cmd.checkout(
                    url=self.url,
                    username=self.username,
                    password=self.password,
                    non_interactive=True,
                    quiet=True,
                    check_returncode=True,
                    **kwargs,
                )
            except exc.CommandError as e:
                result.add_error("checkout", str(e), exception=e)
        else:
            try:
                self.obtain()
            except exc.CommandError as e:
                self.log.exception("Failed to obtain repository")
                result.add_error("obtain", str(e), exception=e)
                return result
            return self.update_repo()
        return result

    @classmethod
    def _get_svn_url_rev(cls, location: str) -> tuple[str | None, int]:
        svn_xml_url_re = re.compile(r'url="([^"]+)"')
        svn_rev_re = re.compile(r'committed-rev="(\d+)"')
        svn_info_xml_rev_re = re.compile(r'\s*revision="(\d+)"')
        svn_info_xml_url_re = re.compile(r"<url>(.*)</url>")

        entries_path = pathlib.Path(location) / ".svn" / "entries"
        if entries_path.exists():
            with entries_path.open() as f:
                data = f.read()
        else:  # subversion >= 1.7 does not have the 'entries' file
            data = ""

        url = None
        if data.startswith(("8", "9", "10")):
            entries = list(map(str.splitlines, data.split("\n\x0c\n")))
            del entries[0][0]  # get rid of the '8'
            url = entries[0][3]
            revs = [int(d[9]) for d in entries if len(d) > 9 and d[9]] + [0]
        elif data.startswith("<?xml"):
            match = svn_xml_url_re.search(data)
            if not match:
                raise SvnUrlRevFormattingError(data=data)
            url = match.group(1)  # get repository URL
            revs = [int(m.group(1)) for m in svn_rev_re.finditer(data)] + [0]
        else:
            try:
                # Note that using get_remote_call_options is not necessary here
                # because `svn info` is being run against a local directory.
                # We don't need to worry about making sure interactive mode
                # is being used to prompt for passwords, because passwords
                # are only potentially needed for remote server requests.
                xml = Svn(path=pathlib.Path(location).parent).info(
                    target=pathlib.Path(location),
                    xml=True,
                )
                match = svn_info_xml_url_re.search(xml)
                assert match is not None
                url = match.group(1)
                revs = [int(m.group(1)) for m in svn_info_xml_rev_re.finditer(xml)]
            except Exception:
                url, revs = None, []

        rev = max(revs) if revs else 0

        return url, rev



"""Exceptions for libvcs."""

from __future__ import annotations


class LibVCSException(Exception):
    """Standard exception raised by libvcs."""


class CommandError(LibVCSException):
    """Raised on non-zero return codes."""

    def __init__(
        self,
        output: str,
        returncode: int | None = None,
        cmd: str | list[str] | None = None,
    ) -> None:
        self.returncode = returncode
        self.output = output
        if cmd:
            if isinstance(cmd, list):
                cmd = " ".join(cmd)
            self.cmd = cmd

    def __str__(self) -> str:
        """Return command output."""
        message = self.message.format(returncode=self.returncode, cmd=self.cmd)
        if len(self.output.strip()):
            message += f"\n{self.output}"
        return message

    message = "Command failed with code {returncode}: {cmd}"


class CommandTimeoutError(CommandError):
    r"""CommandError raised when a subprocess exceeds its wall-clock timeout.

    Carries the deadline used when the timeout fired so callers and human
    readers see the duration in the rendered message rather than the
    misleading ``"Command failed with code None: ..."`` produced by the
    bare :class:`CommandError` template (the child is not necessarily reaped
    before the exception is raised, leaving ``returncode`` ``None``).

    Examples
    --------
    >>> err = CommandTimeoutError(
    ...     output='partial output',
    ...     cmd='git fetch',
    ...     timeout=2.5,
    ... )
    >>> print(str(err))
    Command timed out after 2.5s: git fetch
    partial output
    """

    def __init__(
        self,
        output: str,
        returncode: int | None = None,
        cmd: str | list[str] | None = None,
        timeout: float | None = None,
    ) -> None:
        super().__init__(output=output, returncode=returncode, cmd=cmd)
        #: Wall-clock deadline (seconds) the subprocess exceeded, if known.
        self.timeout = timeout

    def __str__(self) -> str:
        """Render the timeout duration alongside the command and partial output."""
        cmd = getattr(self, "cmd", "")
        if self.timeout is not None:
            message = f"Command timed out after {self.timeout:g}s: {cmd}"
        else:
            message = f"Command timed out: {cmd}"
        if self.output.strip():
            message += f"\n{self.output}"
        return message


class InvalidVCS(LibVCSException):
    """Invalid VCS."""
